@@ -1,4 +1,3 @@
-#ident "$Id: mount_ext2.c,v 1.18 2005/01/10 13:28:29 raven Exp $"
 /* ----------------------------------------------------------------------- *
  *   
  *  mount_ext2.c - module for Linux automountd to mount ext2 filesystems
@@ -16,10 +15,8 @@
 
 #include <stdio.h>
 #include <malloc.h>
-#include <errno.h>
 #include <fcntl.h>
 #include <unistd.h>
-#include <syslog.h>
 #include <string.h>
 #include <stdlib.h>
 #include <sys/param.h>
@@ -38,43 +35,47 @@ int mount_init(void **context)
 	return 0;
 }
 
-int mount_mount(const char *root, const char *name, int name_len,
+int mount_mount(struct autofs_point *ap, const char *root, const char *name, int name_len,
 		const char *what, const char *fstype, const char *options, void *context)
 {
 	char *fullpath;
+	char buf[MAX_ERR_BUF];
 	const char *p, *p1;
 	int err, ro = 0;
 	const char *fsck_prog;
-	int status, existed = 1;
+	int len, status, existed = 1;
 
-	fullpath = alloca(strlen(root) + name_len + 2);
-	if (!fullpath) {
-		error(MODPREFIX "alloca: %m");
-		return 1;
+	if (ap->flags & MOUNT_FLAG_REMOUNT)
+		return 0;
+
+	/* Root offset of multi-mount */
+	len = strlen(root);
+	if (root[len - 1] == '/') {
+		fullpath = alloca(len);
+		len = snprintf(fullpath, len, "%s", root);
+	/* Direct mount name is absolute path so don't use root */
+	} else if (*name == '/') {
+		fullpath = alloca(len + 1);
+		len = sprintf(fullpath, "%s", root);
+	} else {
+		fullpath = alloca(len + name_len + 2);
+		len = sprintf(fullpath, "%s/%s", root, name);
 	}
+	fullpath[len] = '\0';
 
-	if (name_len)
-		sprintf(fullpath, "%s/%s", root, name);
-	else
-		sprintf(fullpath, "%s", root);
-
-	debug(MODPREFIX "calling mkdir_path %s", fullpath);
+	debug(ap->logopt, MODPREFIX "calling mkdir_path %s", fullpath);
 
 	status = mkdir_path(fullpath, 0555);
 	if (status && errno != EEXIST) {
-		error(MODPREFIX "mkdir_path %s failed: %m", fullpath);
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		error(ap->logopt,
+		      MODPREFIX "mkdir_path %s failed: %s", fullpath, estr);
 		return 1;
 	}
 
 	if (!status)
 		existed = 0;
 
-	if (is_mounted(_PATH_MOUNTED, fullpath)) {
-		error(MODPREFIX 
-		  "warning: %s is already mounted", fullpath);
-		return 0;
-	}
-
 	if (options && options[0]) {
 		for (p = options; (p1 = strchr(p, ',')); p = p1)
 			if (!strncmp(p, "ro", p1 - p) && ++p1 - p == sizeof("ro"))
@@ -92,41 +93,55 @@ int mount_mount(const char *root, const char *name, int name_len,
 	fsck_prog = PATH_E2FSCK;
 #endif
 	if (ro) {
-		debug(MODPREFIX "calling %s -n %s", fsck_prog, what);
-		err = spawnl(LOG_DEBUG, fsck_prog, fsck_prog, "-n", what, NULL);
+		debug(ap->logopt,
+		      MODPREFIX "calling %s -n %s", fsck_prog, what);
+		err = spawnl(ap->logopt, fsck_prog, fsck_prog, "-n", what, NULL);
 	} else {
-		debug(MODPREFIX "calling %s -p %s", fsck_prog, what);
-		err = spawnl(LOG_DEBUG, fsck_prog, fsck_prog, "-p", what, NULL);
+		debug(ap->logopt,
+		      MODPREFIX "calling %s -p %s", fsck_prog, what);
+		err = spawnl(ap->logopt, fsck_prog, fsck_prog, "-p", what, NULL);
 	}
 
-	if (err & ~6) {
-		error(MODPREFIX "%s: filesystem needs repair, won't mount",
+	/*
+	 * spawnl returns the error code, left shifted by 8 bits.  We are
+	 * interested in the following error bits from the fsck program:
+	 *    2 - File system errors corrected, system should be rebooted
+	 *    4 - File system errors left uncorrected
+	 */
+	if ((err >> 8) & 6) {
+		error(ap->logopt,
+		      MODPREFIX "%s: filesystem needs repair, won't mount",
 		      what);
 		return 1;
 	}
 
 	if (options) {
-		debug(MODPREFIX "calling mount -t %s " SLOPPY "-o %s %s %s",
+		debug(ap->logopt,
+		      MODPREFIX "calling mount -t %s " SLOPPY "-o %s %s %s",
 		      fstype, options, what, fullpath);
-		err = spawnll(LOG_NOTICE,
-		             PATH_MOUNT, PATH_MOUNT, "-t", fstype,
+		err = spawn_mount(ap->logopt, "-t", fstype,
 			     SLOPPYOPT "-o", options, what, fullpath, NULL);
 	} else {
-		debug(MODPREFIX "calling mount -t %s %s %s",
+		debug(ap->logopt,
+		      MODPREFIX "calling mount -t %s %s %s",
 		      fstype, what, fullpath);
-		err = spawnll(LOG_NOTICE,
-			     PATH_MOUNT, PATH_MOUNT, "-t", fstype,
-			     what, fullpath, NULL);
+		err = spawn_mount(ap->logopt, "-t", fstype, what, fullpath, NULL);
 	}
 
 	if (err) {
-		if ((!ap.ghost && name_len) || !existed)
-			rmdir_path(name);
-		error(MODPREFIX "failed to mount %s (type %s) on %s",
-		      what, fstype, fullpath);
+		info(ap->logopt, MODPREFIX "failed to mount %s (type %s) on %s",
+		    what, fstype, fullpath);
+
+		if (ap->type != LKP_INDIRECT)
+			return 1;
+
+		if ((!(ap->flags & MOUNT_FLAG_GHOST) && name_len) || !existed)
+			rmdir_path(ap, fullpath, ap->dev);
+
 		return 1;
 	} else {
-		debug(MODPREFIX "mounted %s type %s on %s",
+		info(ap->logopt,
+		      MODPREFIX "mounted %s type %s on %s",
 		      what, fstype, fullpath);
 		return 0;
 	}
